!  quedo.f90
!  
!  Copyright 2018 Igor J. Chybicki <igorchy@ukw.edu.pl>
!  
!  This program is free software; you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation; either version 2 of the License, or
!  (at your option) any later version.
!  
!  This program is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!  
!  You should have received a copy of the GNU General Public License
!  along with this program; if not, write to the Free Software
!  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
!  MA 02110-1301, USA.
!  
program quedo
   
   use omp_lib
   use iso_fortran_env
   
   implicit none
   
   !global variables
   integer :: count_rate, count_max, c1, c2, seed, ncor, ninc, usernth, ii
   double precision :: logl,mlogl,vlogl,itm,ie1,ie2,ifk
   double precision :: mltp = dble(10000)
   double precision :: my2pi = dble(2)*dacos(dble(-1.0))
   character(len=258) :: fi,fd,forun,fosum,fomg,forev
   character(len=255) :: domom
   character(len=10) :: LASTUPDATE,VERSION
   character(1) :: tab = char(9)
   character(1) :: dirsep = char(47)            !directory separator; linux mode but works in win as well
   
   !data
   integer :: np,nf,nl,nv,maxl,npop             !number of progeny individuals, families, loci, alleles, populations
   integer, allocatable :: npi(:)               !number of progeny per family
   integer, allocatable :: pgen(:,:,:)          !genotypes of progeny
   integer, allocatable :: mgen(:,:,:)          !genotypes of mother
   double precision, allocatable :: mvar(:,:)   !measured variables for families
   integer, allocatable :: mpop(:)              !population indicator of mother
   integer, allocatable :: popid(:)             !list of population indicators
   integer, allocatable :: pfam(:)              !family indicator for the i-th progeny
   integer, allocatable :: mfam(:)              !family indicator for the i-th mother
   integer, allocatable :: allele(:,:)          !array of alleles
   integer, allocatable :: nall(:)              !number of alleles at the i-th locus
   integer :: cor_dat                           !binary indicator for correcting data
   double precision, allocatable :: pol(:)      !per-locus frequencies of detected parent-offspring incompatibilities
   double precision, allocatable :: ngl(:)      !per-locus frequencies of checked parent-offspring incompatibilities
   double precision, allocatable :: m_var(:),s_var(:) !mean and std. dev. of variables
   
   !missing data (auxiliary variables)
   integer, allocatable :: xi(:),yi(:)          !outcross and selfing event counts per family
   double precision, allocatable :: tj(:)       !outcross rate per individual
   integer, allocatable :: zklm(:,:,:)          !allele counts in pops
   integer, allocatable :: pgjl(:,:)            !inferred pollen gamete
   integer, allocatable :: xj(:),yj(:)          !inferred outcrossing and selfing counts
   integer, allocatable :: initmgen(:,:,:)      !genotypes of mother
   
   !parameters of the model
   double precision, allocatable :: qlm(:,:)    !global allele frequencies
   double precision, allocatable :: pklm(:,:,:) !allele frequencies within the k-th location
   double precision, allocatable :: fk(:)       !divergence rate for the k-th location
   double precision, allocatable :: ti(:)       !family-specific multilocus outcrossing rate
   double precision, allocatable :: e1(:),e2(:) !typing error rates
   integer :: snglloc_t                         !flag (0/1) for indicating if single-locus t is used (multilocus approach otherwise)
   integer :: null_regrmod                      !flag (0/1) for indicating if a constant only present
   integer :: ztrans                            !flag (0/1) for indicating if variables are to be Z-transformed
   integer :: nostruct                          !flag (0/1) for indicating if HW conditions hold for gene pools
   integer :: revjump = 1                       !flag (0/1) for indicating if between-model jumps are performed
   
   !hyper-parameters
   double precision :: yt,mf,kf,me1,ye1,me2,ye2
   
   !hyper-parameters of regression model for t
   double precision :: beta0
   double precision, allocatable :: beta(:)
   integer, allocatable :: y_beta(:)
   double precision, allocatable :: mt(:)
   double precision :: mu_beta,s2_beta
   
   !parameter means across MCMC
   double precision, allocatable :: mti(:)          !family-specific multilocus outcrossing rate
   double precision, allocatable :: m_e1(:),m_e2(:) !typing error rate
   double precision, allocatable :: mpklm(:,:,:)    !lineage allele frequencies
   double precision, allocatable :: mfk(:)          !lineage divergence rate
   
   !mcmc settings
   integer :: nu,bu,pu,ke                       !number of updates (mcmc cycles), thinning (keep every...)
   integer :: upd_e                             !flags (0/1) for treating parameters as estimable
   integer :: infer_mgen                        !flag (0/1) for indicating if maternal gnotypes are estimable
   
   !prior distribution constraints
   double precision :: MINFK,MAXFK
   double precision :: MINQLM,MAXQLM
   double precision :: MINB0,MAXB0
   double precision :: MINYT
   
   !acceptance rates and proposal parameters
   double precision :: v_beta0,ar_beta0,rr_beta0
   double precision, allocatable :: v_beta(:),ar_beta(:),rr_beta(:)
   double precision, allocatable :: mq_beta(:),s2q_beta(:)
   double precision, allocatable :: d_fk(:),ar_fk(:),rr_fk(:)
   double precision, allocatable :: d_qlm(:),ar_qlm(:),rr_qlm(:)
   double precision, allocatable :: d_e1(:),ar_e1(:),rr_e1(:)
   double precision, allocatable :: d_e2(:),ar_e2(:),rr_e2(:)
   double precision :: d_yt,ar_yt,rr_yt
   double precision :: d_mf,ar_mf,rr_mf
   double precision :: d_kf,ar_kf,rr_kf
   double precision :: d_ye1,ar_ye1,rr_ye1
   double precision :: d_me1,ar_me1,rr_me1
   double precision :: d_ye2,ar_ye2,rr_ye2
   double precision :: d_me2,ar_me2,rr_me2
   
   !rng-related variables
   integer :: jsr
   integer, allocatable :: pjsr(:)
   
   ! -------------------------------------------------------------------
   ! SOME INITIALIZATIONS:
   
   !====================================================================
   ! VERSION HISTORY:
   LASTUPDATE='11.02.2019'
   VERSION='0.7'
   !====================================================================
   
   !====================================================================
   ! These constraints seem to work fine:
   MINYT=   dble(0)
   MINFK=   dble(0.000001)
   MAXFK=   dble(0.999999)
   MINQLM=  dble(0.000001)
   MAXQLM=  dble(0.999999)
   MINB0=   dlog(dble(0.000001)/dble(0.999999))
   MAXB0=   -MINB0
   !====================================================================
   
   ! -------------------------------------------------------------------
   ! THE MAIN PROGRAM STARTS HERE:
   
   write(*,"(a)") '---------------------------------------------------------'
   write(*,"(a)") '                          quedo                          '
   write(*,"(a)") '                      version: '//trim(VERSION)
   write(*,"(a)") '                 last update: '//trim(LASTUPDATE)
   write(*,"(a)") '               written by Igor J. Chybicki               '
   write(*,"(a)") '---------------------------------------------------------'
   write(*,*)
   
   fi="info.txt"
   
   call read_info
   call read_data
   
   write(*,"('data file:   ',a)") trim(fd)
   write(*,"('output file: ',a)") trim(forun)
   
   call identify_parents
   call review_data
   call identify_alleles
   call order_alleles
   call identify_pops
   call compute_mean_sd_variables
   call write_rev_data
   
   jsr=seed
   allocate(pjsr(omp_get_max_threads()))
   do ii=1,omp_get_max_threads()
     pjsr(ii)=intuniformdev(100000)+1
   enddo
   
   call compute_npi
   
   write(*,"(a)") '---------------------------------------------------------'
   write(*,"(a,i0)") 'number of pilot cycles:     ',bu+pu
   write(*,"(a,i0)") 'number of sampling cycles:  ',nu
   write(*,"(a,i0)") 'number of saved samples:    ',nu/ke
   write(*,"(a)") '---------------------------------------------------------'
   
   call initialize_mcmc
   
   call system_clock(c1, count_rate, count_max)
   
   call run_mcmc
   
   call system_clock(c2, count_rate, count_max)
   c2=c2-c1
   write(*,"(a)") '---------------------------------------------------------'
   write(*,"('time in sec.: ',i0)") c2/1000
   
 ! THE MAIN PROGRAM ENDS HERE
 
 ! ---------------------- functions & procedures -----------------------
 contains
 
 !----------------------------------------------------------------------
 subroutine read_info
   integer :: infile,unth
   character(len=256) :: n1,n2
   
   infile=44
   cor_dat=0
   open (unit=infile, file=fi) 
   read(infile,*) bu,pu,nu,ke                             !number of updates,keep every...(thinning)
   read(infile,*) snglloc_t,upd_e,infer_mgen,null_regrmod,ztrans,unth !use 0 and 1 to disable/enable parameter updating
   read(infile,*) itm,ifk,ie1,ie2                         !initial parameter values
   read(infile,*) seed                                    !seed to initialize random number generator
   read (infile,*) n1                                     !data file name (max=256 characters!)
   read (infile,*) n2                                     !mcmc output file name (max=256 characters!)
   if (upd_e==0 .and. ie1==dble(0) .and. ie2==dble(0)) cor_dat=1
   fd = n1
   forun = trim(n2)//'.run'
   fosum = trim(n2)//'.sum'
   forev = trim(n2)//'.rev'
   domom = trim(n2)//'_mom'
   if (infer_mgen==1) fomg = trim(n2)
   nostruct= 0
   if (ifk==dble(0)) nostruct= 1
   usernth=omp_get_max_threads()
   if (unth<usernth .and. unth>0) usernth=unth
   
   close(infile)
 end subroutine read_info
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine read_data
   integer :: i,j,k
   
   open (unit=99, file=trim(fd))
   read(99, *) nf,np,nl,nv  !number of progeny, families and markers 
   allocate(pgen(np,nl,2))  !assumes diploidy
   allocate(mgen(nf,nl,2))  !assumes diploidy
   allocate(pfam(np))       !family identifier for each individual
   allocate(mfam(nf))       !family identifier for each individual
   allocate(mpop(nf))       !family identifier for each individual
   allocate(mvar(nf,nv))    !family identifier for each individual
   
   !read data from file
   do i=1,nf
     read(99,*) mfam(i),((mgen(i,j,k),k=1,2),j=1,nl),(mvar(i,j),j=1,nv),mpop(i)
   enddo
   
   do i=1,np
      read(99,*) pfam(i),((pgen(i,j,k),k=1,2),j=1,nl)
   enddo
   close(99)
   
 end subroutine read_data
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine identify_parents
   integer :: i,f
   
   do i=1,np
     do f=1,nf
       if (pfam(i)==mfam(f)) then
         pfam(i)=f
         exit
       endif
     enddo
   enddo
   
 end subroutine identify_parents
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine review_data
   integer :: j,l,m,c,ai,aj,am,an,ntot
   
   ncor=0
   ntot=0
   ninc=0
   allocate(pol(nl))
   allocate(ngl(nl))
   do l=1,nl
     pol(l) = dble(0)
     ngl(l) = dble(0)
   enddo
   
   do j=1,np
       
       m=pfam(j)
       do l=1,nl
         ai=mgen(m,l,1)
         aj=mgen(m,l,2)
         am=pgen(j,l,1)
         an=pgen(j,l,2)
         
         if (am > -1 .and. an > -1) then
           if (ai > -1 .and. aj > -1) then
             ntot=ntot+1
             ngl(l)=ngl(l)+dble(1)
             c = 0
             if (am == ai) then
               c = c + 1
             endif
             if (am == aj) then
               c = c + 1
             endif
             if (an == ai) then
               c = c + 1
             endif
             if (an == aj) then
               c = c + 1
             endif
             
             if (c == 0) then
               if (cor_dat == 1) then
                 pgen(j,l,1) = -1
                 pgen(j,l,2) = -1
                 ncor=ncor+1
               endif
               pol(l)=pol(l)+dble(1)
               ninc=ninc+1
             endif
           endif
         endif
         
       enddo
       
   enddo
   
   write(*,"(a)")                '---------------------------------------------------------'
   write (*,"(a)")               'data review for compatibility:'
   write (*,"(i0,a)")       ntot,' progeny genotypes reviewed'
   write (*,"(i0,a)")       ninc,' parent-offspring incompatibilities detected'
   if (ncor>0) then 
     write (*,"(i0,a)")     ncor,' incompatible genotypes removed'
   endif
   write(*,"(a)")                '---------------------------------------------------------'
 end subroutine review_data
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 integer function cod2all(l,c)
   integer, intent(in) :: l,c
   
   cod2all = -1
   if (c>0) cod2all=allele(l,c)
 end function cod2all
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine write_rev_data
   integer :: i,j,k,fu
   
   fu=851
   open (unit=fu,file=forev,action="write",status="replace")
   write (fu,"(a)") 'Reviewed genotypes:'
   do i=1,nf
     write(fu,"(i0,*(a,i0))") mfam(i),((tab,cod2all(j,mgen(i,j,k)),k=1,2),j=1,nl)
   enddo
   
   do i=1,np
     write(fu,"(i0,*(a,i0))") mfam(pfam(i)),((tab,cod2all(j,pgen(i,j,k)),k=1,2),j=1,nl)
   enddo
   
   close(fu)
   
 end subroutine write_rev_data
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine compute_npi
   integer :: i,n0
   
   allocate(npi(nf))
   do i=1,nf
     npi(i)=0
   enddo
   
   do i=1,np
     npi(pfam(i))=npi(pfam(i))+1
   enddo
   
   n0=0
   do i=1,nf
     if (npi(i)==0) n0=n0+1
   enddo
   
   if (n0>0) then
     write(*,"(a)") '-------------------------------------------------'
     write(*,"(a)") 'warning!'
     write (*,"(i0,a)") n0, ' families without progeny'
   endif
   
 end subroutine compute_npi
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine identify_alleles
   !(double -1 used for missing genotype)
   integer, allocatable ::  tmpallele(:,:)
   integer :: i,j,k,l,m,b
 
   allocate(allele(nl,0))
   allocate(nall(nl))
   allocate(tmpallele(nl,0))
   
   do i=1,nl
     nall(i)=0
   enddo
   
   maxl=0
   
   !alleles in parents
   do i=1,nf
     do j=1,nl
       do k=1,2
         b=0
         do l=1,nall(j)
           if (mgen(i,j,k)==allele(j,l)) then
             b=b+1
           endif
         enddo
         
         if ((mgen(i,j,k).ne.-1).and.(b==0)) then
           tmpallele=allele
           nall(j)=nall(j)+1
           if (maxl<nall(j)) then
             maxl=nall(j)
             deallocate(allele)
             allocate(allele(nl,maxl))
           endif
           do l=1,nl
             do m=1,nall(l)
               if (l.ne.j) then
                 allele(l,m)=tmpallele(l,m)
               else if (m.lt.nall(l)) then
                 allele(l,m)=tmpallele(l,m)
               endif
             enddo
           enddo
           allele(j,nall(j))=mgen(i,j,k)
           deallocate(tmpallele)
           allocate(tmpallele(nl,maxl))
         endif
       enddo
     enddo
   enddo
   
   !alleles in progeny
   do i=1,np
     do j=1,nl
       do k=1,2
         b=0
         do l=1,nall(j)
           if (pgen(i,j,k)==allele(j,l)) then
             b=b+1
           endif
         enddo
         
         if ((pgen(i,j,k).ne.-1).and.(b==0)) then
           tmpallele=allele
           nall(j)=nall(j)+1
           if (maxl<nall(j)) then
             maxl=nall(j)
             deallocate(allele)
             allocate(allele(nl,maxl))
           endif
           do l=1,nl
             do m=1,nall(l)
               if (l.ne.j) then
                 allele(l,m)=tmpallele(l,m)
               else if (m.lt.nall(l)) then
                 allele(l,m)=tmpallele(l,m)
               endif
             enddo
           enddo
           allele(j,nall(j))=pgen(i,j,k)
           deallocate(tmpallele)
           allocate(tmpallele(nl,maxl))
         endif
       enddo
     enddo
   enddo
   
   b=0
   do i=1,nl
     b=b+nall(i)
   enddo
   write(*,"(a, i0)") 'total number of alleles: ',b
   
   !now encode alleles using their indices in the allele table
   do i=1,nf
     do j=1,nl
       do k=1,2
         do m=1,nall(j)
           if (mgen(i,j,k)==allele(j,m)) then
             mgen(i,j,k)=m
             exit
           endif
         enddo
       enddo
     enddo
   enddo

   do i=1,np
     do j=1,nl
       do k=1,2
         do m=1,nall(j)
           if (pgen(i,j,k)==allele(j,m)) then
             pgen(i,j,k)=m
             exit
           endif
         enddo
       enddo
     enddo
   enddo
   
   
   do l=1,nl
     do k=nall(l)+1,ubound(allele,2)
       allele(l,k)=-1
     enddo
   enddo
 end subroutine identify_alleles
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine order_alleles
   integer :: i,j,k
   
   do i=1,nf
     do j=1,nl
       if (mgen(i,j,1) > mgen(i,j,2)) then
         k = mgen(i,j,1)
         mgen(i,j,1) = mgen(i,j,2)
         mgen(i,j,2) = k
       endif
     enddo
   enddo

   do i=1,np
     do j=1,nl
       if (pgen(i,j,1) > pgen(i,j,2)) then
         k = pgen(i,j,1)
         pgen(i,j,1) = pgen(i,j,2)
         pgen(i,j,2) = k
       endif
     enddo
   enddo
 end subroutine order_alleles
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine identify_pops
   integer :: i,b,k
   integer, allocatable :: tmppop(:)
   
   allocate(popid(0))
   allocate(tmppop(0))
   
   npop=0
   do i=1,nf
     b=0
     do k=1,npop
       if (mpop(i)==popid(k)) b=b+1
     enddo
         
     if (b==0) then
       tmppop=popid
       npop=npop+1
       deallocate(popid)
       allocate(popid(npop))
       do k=1,npop-1
         popid(k)=tmppop(k)
       enddo
       popid(npop)=mpop(i)
       deallocate(tmppop)
       allocate(tmppop(npop))
     endif
   enddo
   
   do i=1,nf
     do k=1,npop
       if (mpop(i)==popid(k)) then
         mpop(i)=k
         exit
       endif
     enddo
   enddo
   write(*,"(a, i0)") 'number of populations:   ',npop
   
 end subroutine identify_pops
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine compute_mean_sd_variables
   integer :: i,m
   
   allocate(m_var(nv))
   allocate(s_var(nv))
   
   do m=1,nv
     m_var(m)=dble(0)
     s_var(m)=dble(0)
     do i=1,nf
       m_var(m)=m_var(m)+mvar(i,m)
       s_var(m)=s_var(m)+mvar(i,m)**2
     enddo
     s_var(m)=dsqrt((s_var(m)-m_var(m)*m_var(m)/dble(nf))/dble(nf-1))
     m_var(m)=m_var(m)/dble(nf)
     
     if (s_var(m)>dble(0) .and. ztrans==1) then
       do i=1,nf
         mvar(i,m)=(mvar(i,m)-m_var(m))/s_var(m)
       enddo
     endif
   enddo
   
 end subroutine compute_mean_sd_variables
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine init_maternal_genotypes
   integer :: i,j,k,lo,li,m,n,ai,aj,am,an,na,mmax,nmax,combimpos
   double precision :: pr,lc,lmax
   
   write(*,"(a)",advance="no")                'initial search for maternal genotypes: '
   allocate(initmgen(nf,nl,2))
  
   do i=1,nf
     k=mpop(i)
     do lo=1,nl
       lmax = -1.D300
       mmax = 0
       nmax = 0
       
       do m=1,nall(lo)
         initmgen(i,lo,1)=m
         
         do n=m,nall(lo)
           initmgen(i,lo,2)=n
           combimpos=0
           lc=dble(0)
           do j=1,np
             if (pfam(j).ne.i) cycle
             Pr = dble(1)
             if (combimpos==0) then
               do li=1,lo
                 ai=initmgen(i,li,1)
                 aj=initmgen(i,li,2)
                 am=pgen(j,li,1)
                 an=pgen(j,li,2)
                 na=nall(li)
                 if (am>-1 .and. an>-1 .and. pr>dble(0)) then
                   pr = pr * pr_tt(ai,aj,am,an,k,li,na,e1(li),e2(li))
                 endif
               enddo
               if (pr > dble(0)) then
                 lc=lc+dlog(pr)
               else 
                 combimpos=1
               endif
             endif
           enddo
           
           if (lc > lmax .and. combimpos==0) then
             mmax = initmgen(i,lo,1)
             nmax = initmgen(i,lo,2)
             lmax = lc
           endif
         enddo
       enddo
       initmgen(i,lo,1)= mmax
       initmgen(i,lo,2)= nmax
     enddo
     
     do lo=1,nl
       if (initmgen(i,lo,1)<initmgen(i,lo,2)) then
         mgen(i,lo,1)=initmgen(i,lo,1)
         mgen(i,lo,2)=initmgen(i,lo,2)
       else
         mgen(i,lo,1)=initmgen(i,lo,2)
         mgen(i,lo,2)=initmgen(i,lo,1)
       endif
     enddo
   enddo
   call write_rev_data
   write(*,"(a)",advance="yes")  'done'
   write(*,"(a)")                '---------------------------------------------------------'
   
 end subroutine init_maternal_genotypes
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function uniformdev()
   uniformdev=uniziggurat()
 end function uniformdev
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function puniformdev(ujsr)
   integer, intent(inout) :: ujsr
   
   puniformdev=puniziggurat(ujsr)
 end function puniformdev
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 integer function intuniformdev(n)
   integer, intent(in) :: n
 
   intuniformdev = int(uniformdev() * dble(n))
 end function intuniformdev
 !----------------------------------------------------------------------
  
 !----------------------------------------------------------------------
 integer function pintuniformdev(n,ujsr)
   integer, intent(in) :: n
   integer, intent(inout) :: ujsr
 
   pintuniformdev = int(puniformdev(ujsr) * dble(n))
 end function pintuniformdev
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function gammadev(a)
   double precision, intent(in) :: a
  
   if (a<dble(1)) then
     gammadev=rgamma(a+dble(1))*uniformdev()**(dble(1)/a)
   else
     gammadev=rgamma(a)
   endif
  
 end function gammadev
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function pgammadev(a,ujsr)
   double precision, intent(in) :: a
   integer, intent(inout) :: ujsr
  
   if (a<dble(1)) then
     pgammadev=prgamma(a+dble(1),ujsr)*puniformdev(ujsr)**(dble(1)/a)
   else
     pgammadev=prgamma(a,ujsr)
   endif
  
 end function pgammadev
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function normaldev(m,s)
   double precision, intent(in) :: m,s
   double precision :: v1,v2,r
   
   do
     v1 = dble(2) * uniformdev() - dble(1)
     v2 = dble(2) * uniformdev() - dble(1)
     r  = v1**2 + v2**2
     if (r < dble(1) .and. r > dble(0)) exit
   enddo
   normaldev = v2 * dsqrt(-dble(2) * dlog(r) / r) * s + m
   
 end function normaldev
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function rectdev(x,d)
   !returns a new value
   !using the rectangular distr. reflection
   double precision, intent(in) :: x,d
   double precision :: nx
   
   nx = x + dble(dble(0.5) - uniformdev()) * d
   if (nx < dble(0)) then
     nx = -nx
   else
     if (nx > dble(1)) then
       nx = dble(2) - nx
     endif
   endif
   
   rectdev = nx
 end function rectdev
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function prectdev(x,d,ujsr)
   !returns a new value
   !using the rectangular distr. reflection
   double precision, intent(in) :: x,d
   integer, intent(inout) :: ujsr
   double precision :: nx
   
   nx = x + dble(dble(0.5) - puniformdev(ujsr)) * d
   if (nx < dble(0)) then
     nx = -nx
   else
     if (nx > dble(1)) then
       nx = dble(2) - nx
     endif
   endif
   
   prectdev = nx
 end function prectdev
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function pr_rg(uu,vv,ww,xx,k,ep1,ep2)
   integer, intent(in) :: uu,vv,ww,xx,k
   double precision, intent(in) :: ep1,ep2
   integer :: duv,u,v,w,x,a
   double precision :: e1,e2,pr

   pr = 0.0d0
   
   u = uu
   v = vv
   w = ww
   x = xx
   
   if (u > v) then
     a = u
     u = v
     v = a
   end if
   
   if (w > x) then
     a = w
     w = x
     x = a
   end if
   
   e1 = ep1 / (dble(1) + ep1)
   e2 = ep2 / dble(k - 1)

   if (u == v) then
     duv = 1 
   else
     duv = 0
   endif
   
   if (w /= x) then
     if (u == w .and. v == x) then
       pr = (dble(1) - ep2)**2 + e2**2 - dble(2) * e1 * (dble(1) - ep2 - e2)**2
     else
       if ((u == v .and. u == w) .or. (u == v .and. u == x)) then
         pr = e2 * (dble(1) - ep2) + e1 * (dble(1) - ep2 - e2)**2
       else
         if ((u /= w .and. u /= x) .and. (v /= w .and. v /= x)) then
           pr = dble(2 - duv) * e2**2
         else
           pr = e2 * (dble(1) - ep2 + e2)
         endif
       endif
     endif
   else
     if (u == v .and. u == w) then
       pr = (dble(1) - ep2)**2
     else
       if ((u == w .and. v /= w) .or. (v == w .and. u /= w)) then
         pr = dble(2) * e2 * (dble(1) - ep2)
       else
         if (u /= w .and. v /= w) then
           pr = dble(2 - duv) * e2**2
         endif
       endif
     endif
   endif

   pr_rg = pr
 end function pr_rg
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function pr_s(u,v,w,x,k,ep1,ep2)
   integer, intent(in) :: u,v,w,x,k
   double precision, intent(in) :: ep1,ep2
   
   if (w /= x) then
     pr_s = (pr_rg(u,v,w,w,k,ep1,ep2) + dble(2) * pr_rg(u,v,w,x,k,ep1,ep2) &
     + pr_rg(u,v,x,x,k,ep1,ep2)) * dble(0.25)
   else 
     pr_s = pr_rg(u,v,w,w,k,ep1,ep2)
   endif
 end function pr_s
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function pr_t(u,v,w,x,k,l,n,ep1,ep2)
   integer, intent(in) :: u,v,w,x,k,l,n
   double precision, intent(in) :: ep1,ep2
   integer :: y
   double precision :: spr,pry(n)
   
   spr = 0.0d0
   if (w /= x) then
     do y=1,n
       pry(y) = pklm(k,l,y) * (pr_rg(u,v,w,y,n,ep1,ep2) + pr_rg(u,v,x,y,n,ep1,ep2)) * 0.5d0
       spr = spr + pry(y)
     enddo
   else
     do y=1,n
       pry(y) = pklm(k,l,y) * pr_rg(u,v,w,y,n,ep1,ep2)
       spr = spr + pry(y)
     enddo
   endif
   
   pr_t = spr
   
 end function pr_t
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function pr_ta(u,v,w,x,k,l,na,ep1,ep2,pa,ujsr)
   integer, intent(in) :: u,v,w,x,k,l,na
   double precision, intent(in) :: ep1,ep2
   integer, intent(inout) :: pa,ujsr
   integer :: y
   double precision :: cpr,spr,r,pry(na)
   
   pa = 1
   spr = 0.0d0
   if (w /= x) then
     do y=1,na
       pry(y) = pklm(k,l,y) * (pr_rg(u,v,w,y,na,ep1,ep2) + pr_rg(u,v,x,y,na,ep1,ep2)) * 0.5d0
       spr = spr + pry(y)
     enddo
   else
     do y=1,na
       pry(y) = pklm(k,l,y) * pr_rg(u,v,w,y,na,ep1,ep2)
       spr = spr + pry(y)
     enddo
   endif
   
   pr_ta = spr
   
   cpr = 0.0d0
   r = puniformdev(ujsr)
   do y=1,na
     cpr = cpr + pry(y) / spr
     if (r >= cpr - pry(y) / spr .and. r < cpr) then
       pa = y
       exit
     endif
   enddo
   
 end function pr_ta
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function pr_b(u,v,k,l,na,ep1,ep2)
   integer, intent(in) :: u,v,k,l,na
   double precision, intent(in) :: ep1,ep2
   integer :: y,z,dyz
   double precision :: pr
   
   pr = 0.0d0
   do y=1,na
     do z=y,na
       if (y == z) then
         dyz = 1
       else 
         dyz = 0
       endif
       pr = pr + dble(2 - dyz) * pklm(k,l,y) * pklm(k,l,z) * pr_rg(u,v,y,z,na,ep1,ep2)
     enddo
   enddo
 pr_b = pr
 end function pr_b
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function pr_ts(i,j,m,n,na,ep1,ep2)
   integer, intent(in) :: i,j,m,n,na
   double precision, intent(in) :: ep1,ep2
   double precision :: d_mn,d_im,d_in,d_jm,d_jn,d_ij_m,d_ij_n
   
   if (na < 2) then
     pr_ts = dble(1)
     return
   endif
   
   if (ep1 > dble(0) .or. ep2 > dble(0)) then
     pr_ts=pr_s(m,n,i,j,na,ep1,ep2)
   else
     if (m == n) then
       d_mn = dble(1)
     else
       d_mn = dble(0)
     endif
     
     if (i == m) then
       d_im = dble(1)
     else
       d_im = dble(0)
     endif
     
     if (i == n) then
       d_in = dble(1)
     else
       d_in = dble(0)
     endif
     
     if (j == m) then
       d_jm = dble(1)
     else
       d_jm = dble(0)
     endif
     
     if (j == n) then
       d_jn = dble(1)
     else
       d_jn = dble(0)
     endif
     
     d_ij_m = (d_im + d_jm) / dble(2)
     d_ij_n = (d_in + d_jn) / dble(2)   
     pr_ts = (dble(2) - d_mn) * d_ij_m * d_ij_n
   endif
   
 end function pr_ts
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function pr_tb(m,n,k,l,na,ep1,ep2)
   integer, intent(in) :: m,n,k,l,na
   double precision, intent(in) :: ep1,ep2
   double precision :: d_mn,d_bm,d_bn
   
   if (na < 2) then
     pr_tb = dble(1)
     return
   endif
   
   if (ep1 > dble(0) .or. ep2 > dble(0)) then
     pr_tb = pr_b(m,n,k,l,na,ep1,ep2)
   else
     if (m == n) then
       d_mn = dble(1)
     else
       d_mn = dble(0)
     endif
     
     d_bm = pklm(k,l,m)
     d_bn = pklm(k,l,n)
     pr_tb = (dble(2) - d_mn) * d_bm * d_bn  
   endif
      
 end function pr_tb
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function pr_tt(i,j,m,n,k,l,na,ep1,ep2)
   integer, intent(in) :: i,j,m,n,k,l,na
   double precision, intent(in) :: ep1,ep2
   double precision :: d_mn,d_im,d_in,d_jm,d_jn,d_ij_m,d_ij_n,d_b_m,d_b_n
   
   if (na < 2) then
     pr_tt = dble(1)
     return
   endif
   
   if (ep1 > dble(0) .or. ep2 > dble(0)) then
     pr_tt = pr_t(m,n,i,j,k,l,na,ep1,ep2)
   else
     if (m == n) then 
       d_mn = dble(1)
     else
       d_mn = dble(0)
     endif
     
     if (i == m) then 
       d_im = dble(1)
     else
       d_im = dble(0)
     endif
     
     if (i == n) then 
       d_in = dble(1)
     else
       d_in = dble(0)
     endif
     
     if (j == m) then 
       d_jm = dble(1)
     else
       d_jm = dble(0)
     endif
     
     if (j == n) then 
       d_jn = dble(1)
     else
       d_jn = dble(0)
     endif
     
     d_ij_m = (d_im + d_jm) / dble(2)
     d_ij_n = (d_in + d_jn) / dble(2)

     d_b_m = pklm(k,l,m)
     d_b_n = pklm(k,l,n)
     
     pr_tt = (dble(2) - d_mn) * (d_ij_m * d_b_n + d_ij_n * d_b_m) / dble(2)
     
   endif
 
 end function pr_tt
 !----------------------------------------------------------------------
 
 
 !----------------------------------------------------------------------
 double precision function pr_tta(i,j,m,n,k,l,na,ep1,ep2,pa,ujsr)
   integer, intent(in) :: i,j,m,n,k,l,na
   integer, intent(out) :: pa,ujsr
   double precision, intent(in) :: ep1,ep2
   double precision :: d_mn,d_im,d_in,d_jm,d_jn,d_ij_m,d_ij_n,d_b_m,d_b_n
   
   if (na < 2) then
     pr_tta = dble(1)
     pa = m
     return
   endif
   
   if (ep1 > dble(0) .or. ep2 > dble(0)) then
     pr_tta = pr_ta(m,n,i,j,k,l,na,ep1,ep2,pa,ujsr)
   else
     if (m == n) then 
       d_mn = dble(1)
     else
       d_mn = dble(0)
     endif
     
     if (i == m) then 
       d_im = dble(1)
     else
       d_im = dble(0)
     endif
     
     if (i == n) then 
       d_in = dble(1)
     else
       d_in = dble(0)
     endif
     
     if (j == m) then 
       d_jm = dble(1)
     else
       d_jm = dble(0)
     endif
     
     if (j == n) then 
       d_jn = dble(1)
     else
       d_jn = dble(0)
     endif
     
     d_ij_m = (d_im + d_jm) / dble(2)
     d_ij_n = (d_in + d_jn) / dble(2)

     d_b_m = pklm(k,l,m)
     d_b_n = pklm(k,l,n)
     
     pr_tta = (dble(2) - d_mn) * (d_ij_m * d_b_n + d_ij_n * d_b_m) / dble(2)
     
     if (puniformdev(ujsr) < d_ij_m * d_b_n/(d_ij_m * d_b_n + d_ij_n * d_b_m)) then
       pa = n
     else
       pa = m
     endif
   endif
 
 end function pr_tta
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function lik_m(i)
   integer, intent(in) :: i
   integer :: n,k,l,ai,aj
   double precision :: pr
   
   pr=dble(1)
   k=mpop(i)
   do l=1,nl
     ai=mgen(i,l,1)
     aj=mgen(i,l,2)
     n=nall(l)
     if (ai > -1 .and. aj > -1) then
       pr=pr+dlog(pr_tb(ai,aj,k,l,n,e1(l),e2(l)))
     endif
   enddo
   
   lik_m=pr
   
 end function lik_m
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function lik_m_l(i,l)
   integer, intent(in) :: i,l
   integer :: n,k,ai,aj
   double precision :: pr
   
   pr=dble(1)
   k=mpop(i)
   ai=mgen(i,l,1)
   aj=mgen(i,l,2)
   n=nall(l)
   if (ai > -1 .and. aj > -1) then
     pr=pr_tb(ai,aj,k,l,n,e1(l),e2(l))
   endif
   
   lik_m_l=dlog(pr)
   
 end function lik_m_l
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function lik_p_tm(j)
   integer, intent(in) :: j
   integer :: i,l,k,na,ai,aj,am,an
   double precision :: prs,prt,pr
   
   i=pfam(j)
   k=mpop(i)
   prs=mltp
   prt=mltp

   do l=1,nl
     ai=mgen(i,l,1)
     aj=mgen(i,l,2)
     am=pgen(j,l,1)
     an=pgen(j,l,2)
     na=nall(l)
     
     if (am > -1 .and. an > -1) then
       if (ai > -1 .and. aj > -1) then
         prs = prs * pr_ts(ai,aj,am,an,na,e1(l),e2(l))
         prt = prt * pr_tt(ai,aj,am,an,k,l,na,e1(l),e2(l))
       else
         prs = prs * pr_tb(am,an,k,l,na,e1(l),e2(l))
         prt = prt * pr_tb(am,an,k,l,na,e1(l),e2(l))
       endif
     endif
   enddo
   
   pr=((dble(1) - ti(i)) * prs + ti(i) * prt)
   
   lik_p_tm = dlog(pr)-dlog(mltp)
 end function lik_p_tm
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function lik_p_ts(j)
   integer, intent(in) :: j
   integer :: i,l,n,k,ai,aj,am,an
   double precision :: prs,prt,pr
   
   i=pfam(j)
   k=mpop(i)
   pr=1.0d0
   do l=1,nl
     ai=mgen(i,l,1)
     aj=mgen(i,l,2)
     am=pgen(j,l,1)
     an=pgen(j,l,2)
     n=nall(l)
     
     prs=dble(1)
     prt=dble(1)
     
     if (am > -1 .and. an > -1) then
       if (ai > -1 .and. aj > -1) then
         prs = pr_ts(ai,aj,am,an,n,e1(l),e2(l))
         prt = pr_tt(ai,aj,am,an,k,l,n,e1(l),e2(l))
       else
         prs = pr_tb(am,an,k,l,n,e1(l),e2(l))
         prt = prs
       endif
     endif
     
     pr=pr+dlog((dble(1)-ti(i))*prs+ti(i)*prt)
   enddo
   
   lik_p_ts = pr
 end function lik_p_ts
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function lik_p_ts_l(j,l)
   integer, intent(in) :: j,l
   integer :: i,n,k,ai,aj,am,an
   double precision :: prs,prt,pr
   
   i=pfam(j)
   k=mpop(i)
   pr=1.0d0

     ai=mgen(i,l,1)
     aj=mgen(i,l,2)
     am=pgen(j,l,1)
     an=pgen(j,l,2)
     n=nall(l)
     
     prs=dble(1)
     prt=dble(1)
     
     if (am > -1 .and. an > -1) then
       if (ai > -1 .and. aj > -1) then
         prs = pr_ts(ai,aj,am,an,n,e1(l),e2(l))
         prt = pr_tt(ai,aj,am,an,k,l,n,e1(l),e2(l))
       else
         prs = pr_tb(am,an,k,l,n,e1(l),e2(l))
         prt = prs
       endif
     endif
     
     pr=pr*(dble(1)-ti(i))*prs+ti(i)*prt

   
   lik_p_ts_l = dlog(pr)
 end function lik_p_ts_l
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine infer_aux_ts_j(j,updt_tj,ujsr)
   integer, intent(in) :: j,updt_tj
   integer, intent(inout) :: ujsr
   integer :: i,l,k,n,ai,aj,am,an
   double precision :: pr
   double precision, allocatable :: prs(:),prt(:)
   
   allocate(prs(nl))
   allocate(prt(nl))
   
   i=pfam(j)
   k=mpop(i)
   do l=1,nl
     ai=mgen(i,l,1)
     aj=mgen(i,l,2)
     am=pgen(j,l,1)
     an=pgen(j,l,2)
     n=nall(l)
     pgjl(j,l)=-1
     prs(l)=1.0d0
     prt(l)=1.0d0
     
     if (am > -1 .and. an > -1) then
       if (ai > -1 .and. aj > -1) then
         prs(l) = pr_ts(ai,aj,am,an,n,e1(l),e2(l))
         prt(l) = pr_tta(ai,aj,am,an,k,l,n,e1(l),e2(l),pgjl(j,l),ujsr)
       else
         prs(l) = pr_tb(am,an,k,l,n,e1(l),e2(l))
         prt(l) = prs(l)
         
         if (puniformdev(ujsr) < 0.5) then
           pgjl(j,l)=am
         else
           pgjl(j,l)=an
         endif
         
       endif
     endif
     
   enddo
   
   xj(j)=0
   yj(j)=0
   do l=1,nl
     pr=((dble(1) - ti(i)) * prs(l) + ti(i) * prt(l))
       
     if (puniformdev(ujsr) < (dble(1) - ti(i)) * prs(l) / pr) then
       yj(j)=yj(j)+1
       pgjl(j,l)=-1
     else
       xj(j)=xj(j)+1
     endif
   enddo
   
   if (updt_tj==1) then
     if (xj(j)+yj(j)>0) then
       tj(j)=tj(j)+dble(xj(j))/dble(xj(j)+yj(j))
     else
       tj(j)=tj(j)+dble(0.5)
     endif
   endif
   
   
 end subroutine infer_aux_ts_j
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine infer_aux_tm_j(j,updt_tj,ujsr)
   integer, intent(in) :: j,updt_tj
   integer, intent(inout) :: ujsr
   integer :: i,l,k,n,ai,aj,am,an
   double precision :: pr, prs,prt
   
   i=pfam(j)
   k=mpop(i)
   prs=1.d0
   prt=1.d0
   do l=1,nl
     ai=mgen(i,l,1)
     aj=mgen(i,l,2)
     am=pgen(j,l,1)
     an=pgen(j,l,2)
     n=nall(l)
     pgjl(j,l)=-1
     
     if (am > -1 .and. an > -1) then
       if (ai > -1 .and. aj > -1) then
         prs = prs * pr_ts(ai,aj,am,an,n,e1(l),e2(l))
         prt = prt * pr_tta(ai,aj,am,an,k,l,n,e1(l),e2(l),pgjl(j,l),ujsr)
       else
         pr = pr_tb(am,an,k,l,n,e1(l),e2(l))
         prs = prs * pr
         prt = prs * pr
         
         if (puniformdev(ujsr) < 0.5) then
           pgjl(j,l)=am
         else
           pgjl(j,l)=an
         endif
         
       endif
     endif
     
   enddo
   
   xj(j)=0
   yj(j)=0
   
   pr=(dble(1)-ti(i))*prs+ti(i)*prt
   if (puniformdev(ujsr) < (dble(1)-ti(i))*prs/pr) then
     yj(j)=yj(j)+1
     do l=1,nl
       pgjl(j,l)=-1
     enddo
   else
     xj(j)=xj(j)+1
   endif
   
   if (updt_tj==1) then
     if (xj(j)+yj(j)>0) then
       tj(j)=tj(j)+dble(xj(j))/dble(xj(j)+yj(j))
     else
       tj(j)=tj(j)+dble(0.5)
     endif
   endif
   
 end subroutine infer_aux_tm_j
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine infer_aux(updt_tj)
   integer, intent(in) :: updt_tj
   integer :: i,j,l,k,ai,aj,th,tjsr
   
   
   do i=1,nf
     xi(i)=0
     yi(i)=0
   enddo
   
   do k=1,npop
     do l=1,nl
       do i=1,nall(l)
         zklm(k,l,i)=0
       enddo
     enddo
   enddo
   
   if (snglloc_t==1) then
     !$omp parallel num_threads(usernth) default(shared) private(th,tjsr)
     !$omp do
     do j=1,np
       th=omp_get_thread_num()+1
       tjsr=pjsr(th)
       call infer_aux_ts_j(j,updt_tj,tjsr)
       pjsr(th)=tjsr
     enddo
     !$omp end do
     !$omp end parallel
   else
     !$omp parallel num_threads(usernth) default(shared) private(th,tjsr)
     !$omp do
     do j=1,np
       th=omp_get_thread_num()+1
       tjsr=pjsr(th)
       call infer_aux_tm_j(j,updt_tj,tjsr)
       pjsr(th)=tjsr
     enddo
     !$omp end do
     !$omp end parallel
   endif
   
   do j=1,np
     do l=1,nl
       k=mpop(pfam(j))
       if (pgjl(j,l)>0) zklm(k,l,pgjl(j,l))=zklm(k,l,pgjl(j,l))+1
     enddo
     
     xi(pfam(j))=xi(pfam(j))+xj(j)
     yi(pfam(j))=yi(pfam(j))+yj(j)
   enddo
   
   do i=1,nf
     do l=1,nl
       ai=mgen(i,l,1)
       aj=mgen(i,l,2)
       k=mpop(i)
       if (ai > -1) zklm(k,l,ai)=zklm(k,l,ai)+1
       if (aj > -1) zklm(k,l,aj)=zklm(k,l,aj)+1
     enddo
   enddo
   
 end subroutine infer_aux
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function full_logl()
   double precision :: sli
   integer :: i
   double precision, allocatable :: li(:)
   
   allocate(li(np+nf))
   sli=dble(0)
   
   if (snglloc_t==1) then
     !$omp parallel num_threads(usernth) default(shared)
     !$omp do
     do i=1,np
       li(i)=lik_p_ts(i)
     enddo
     !$omp end do
     !$omp end parallel
   else
     !$omp parallel num_threads(usernth) default(shared)
     !$omp do
     do i=1,np
       li(i)=lik_p_tm(i)
     enddo
     !$omp end do
     !$omp end parallel
   endif
   
   !$omp parallel num_threads(usernth) default(shared)
   !$omp do
   do i=1,nf
     li(i+np)=lik_m(i)
   enddo
   !$omp end do
   !$omp end parallel
   
   do i=1,np+nf
     sli=sli+li(i)
   enddo
   
   full_logl = sli
 end function full_logl
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function logl_ts_l(l)
   integer, intent(in) :: l
   double precision :: sli
   integer :: i
   
   sli=dble(0)
   
   do i=1,np
     sli=sli+lik_p_ts_l(i,l)
   enddo
   
   do i=1,nf
     sli=sli+lik_m_l(i,l)
   enddo
   
   logl_ts_l = sli
 end function logl_ts_l
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function log_BetaBinD(n1,n2,m,y)
   integer, intent(in) :: n1,n2
   double precision, intent(in) :: m,y
   double precision :: a,b
   ! returns logPr up to a constant
   a = m * (dble(1) - y) / y
   b = (dble(1) - m) * (dble(1) - y) / y
   log_BetaBinD = DLGAMA(dble(n1)+a)+DLGAMA(dble(n2)+b)-DLGAMA(dble(n1+n2)+a+b)+DLGAMA(a+b)-DLGAMA(a)-DLGAMA(b)
   
 end function log_BetaBinD
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function log_DirichletD(x,b,n)
   integer, intent(in) :: n
   double precision, intent(in) :: x(n),b(n)
   integer :: i
   double precision :: pg,sb,p
   
   pg=dble(0)
   sb=dble(0)
   p=dble(0)
   do i=1,n
     sb=sb+b(i)
     pg=pg+DLGAMA(b(i))
     p=p+(b(i)-dble(1))*dlog(x(i))
   enddo
   log_DirichletD=p-pg+DLGAMA(sb)
   
 end function log_DirichletD
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function log_GammaD(x,m,k)
   !defined in terms of mean and shape
   double precision, intent(in) :: x,m,k
   double precision :: a,b
   
   a=k
   b=k/m
   log_GammaD=a*dlog(b)-DLGAMA(a)+(a-dble(1))*dlog(x)-b*x
   
 end function log_GammaD
 !----------------------------------------------------------------------
 
 double precision function log_NormD(x,m,v)
   double precision, intent(in) :: x,m,v
   
   log_NormD = -(x-m)**2/(dble(2)*v)-dlog(my2pi*v)*dble(0.5)
 end function log_NormD
 
 !----------------------------------------------------------------------
 double precision function logl_t()
   integer :: i
   double precision :: lik
   
   lik = dble(0)
   call compute_mt

   do i=1,nf
     lik=lik+log_BetaBinD(xi(i),yi(i),mt(i),yt)
   enddo

 logl_t = lik
 end function logl_t
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine compute_mt
   integer :: i,m
   
   do i=1,nf
     mt(i)=beta0
     if (null_regrmod==0) then
       do m=1,nv
         if (y_beta(m)==1) mt(i)=mt(i)+beta(m)*mvar(i,m)
       enddo
      endif
     mt(i)=dble(1)/(dble(1)+dexp(-mt(i)))
   enddo
   
 end subroutine compute_mt
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine draw_Mik(i,ujsr)
   integer, intent(in) :: i
   integer, intent(inout) :: ujsr
   integer :: j,k,l,m,n
   integer :: ai,aj,am,an
   integer, dimension(2) :: M_o
   double precision :: prs_o,prt_o,prs_n,prt_n,ln,lo,prMo,PrMn,pjo,pjn
   
   l=pintuniformdev(nl,ujsr)+1
   m=pintuniformdev(2,ujsr)+1
   if (nall(l)==1) then
     return
   endif
   k=mpop(i)
   M_o(1)=mgen(i,l,1)
   M_o(2)=mgen(i,l,2)
   do
     mgen(i,l,m)=pintuniformdev(nall(l),ujsr)+1
     if (mgen(i,l,m).ne.M_o(m)) exit
   enddo
   
   ln = dble(0)
   lo = dble(0)
   
   do j=1,np
     if (pfam(j)==i) then
       am=pgen(j,l,1)
       an=pgen(j,l,2)
       n=nall(l)
       prs_o=1.d0
       prt_o=1.d0
       prs_n=1.d0
       prt_n=1.d0
       
       if (am > -1 .and. an > -1) then
         if (ai > -1 .and. aj > -1) then
           ai=mgen(i,l,1)
           aj=mgen(i,l,2)
           prs_n = pr_ts(ai,aj,am,an,n,e1(l),e2(l))
           prt_n = pr_tt(ai,aj,am,an,k,l,n,e1(l),e2(l))
           
           ai=M_o(1)
           aj=M_o(2)
           prs_o = pr_ts(ai,aj,am,an,n,e1(l),e2(l))
           prt_o = pr_tt(ai,aj,am,an,k,l,n,e1(l),e2(l))
         else
           ai=mgen(i,l,1)
           aj=mgen(i,l,2)
           prs_n = pr_tb(am,an,k,l,n,e1(l),e2(l))
           prt_n = prs_n
           
           ai=M_o(1)
           aj=M_o(2)
           prs_o = pr_tb(am,an,k,l,n,e1(l),e2(l))
           prt_o = prs_o
         endif
       endif
       
       pjn=ti(i)*prt_n+(dble(1)-ti(i))*prs_n
       pjo=ti(i)*prt_o+(dble(1)-ti(i))*prs_o
       if (pjn>dble(0) .and. pjo>dble(0)) then
         ln = ln + dlog(pjn)
         lo = lo + dlog(pjo)
       else
         mgen(i,l,1)=m_o(1)
         mgen(i,l,2)=m_o(2)
         return
       endif
     endif
   enddo
   
   if (ln==dble(0) .or. lo==dble(0)) return
   
   ai=M_o(1)
   aj=M_o(2)
   PrMo=dlog(pr_tb(ai,aj,k,l,n,dble(0),dble(0)))
   
   ai=mgen(i,l,1)
   aj=mgen(i,l,2)
   PrMn=dlog(pr_tb(ai,aj,k,l,n,dble(0),dble(0)))
   
   if (dlog(puniformdev(ujsr)).ge.(ln - lo + prMn - prMo)) then
     if (m_o(1)<m_o(2)) then
       mgen(i,l,1)=m_o(1)
       mgen(i,l,2)=m_o(2)
     else
       mgen(i,l,1)=m_o(2)
       mgen(i,l,2)=m_o(1)
     endif
   else
     if (ai>aj) then
       mgen(i,l,1)=aj
       mgen(i,l,2)=ai
     endif
   endif
   
 end subroutine draw_Mik
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine draw_ti(j,ujsr)
   integer, intent(in) :: j
   integer, intent(inout) :: ujsr
   integer :: i,sx,sy,n
   double precision, allocatable :: b(:),x(:)
   
   allocate(b(2))
   allocate(x(2))

   if (j>0) then
     b(1)=mt(j)*(dble(1)-yt)/yt
     b(2)=(dble(1)-mt(j))*(dble(1)-yt)/yt
     n=0
     do
       x(1)=pgammadev(dble(xi(j))+b(1),ujsr)
       x(2)=pgammadev(dble(yi(j))+b(2),ujsr)
       ti(j)=x(1)/(x(1)+x(2))
       n=n+1
       if (n>100) ti(j)=mt(j)
       if (ti(j)>dble(0) .and. ti(j)<dble(1)) exit
     enddo
   else
     sx=0
     sy=0
     do i=1,nf
       sx=sx+xi(i)
       sy=sy+yi(i)
     enddo
     
     b(1)=dble(1)
     b(2)=dble(1)
     x(1)=pgammadev(dble(sx)+b(1),ujsr)
     x(2)=pgammadev(dble(sy)+b(2),ujsr)
     
     do i=1,nf
       ti(i)=x(1)/(x(1)+x(2))
     enddo
   endif
  end subroutine draw_ti
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine draw_e1(l,ujsr)
   integer, intent(in) :: l
   integer, intent(inout) :: ujsr
   double precision :: lo,xo,pro,prn,llo,lln
   double precision, allocatable :: b(:),x(:)
   integer :: u
   
   u=0
   
   xo = e1(l)
   lo = logl
   
   llo = logl_ts_l(l)
   
   allocate(b(2))
   allocate(x(2))
   
   b(1)=me1*(dble(1)-ye1)/ye1
   b(2)=(dble(1)-me1)*(dble(1)-ye1)/ye1

   x(1)=e1(l)
   x(2)=dble(1)-e1(l)
   pro = log_DirichletD(x,b,2)
   
   e1(l) = prectdev(xo,d_e1(l),ujsr)
   lln = logl_ts_l(l)
   
   x(1)=e1(l)
   x(2)=dble(1)-e1(l)
   prn = log_DirichletD(x,b,2)
   
   if (dlog(puniformdev(ujsr)).ge.(lln - llo + prn - pro)) then
     e1(l) = xo
     rr_e1(l)=rr_e1(l)+dble(1)
   else
     ar_e1(l)=ar_e1(l)+dble(1)
   endif
   
 end subroutine draw_e1
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine draw_e2(l,ujsr)
   integer, intent(in) :: l
   integer, intent(inout) :: ujsr
   double precision :: lo,xo,pro,prn,llo,lln
   double precision, allocatable :: b(:),x(:)
   integer :: u
   
   u=0
   
   xo = e2(l)
   lo = logl
   
   llo = logl_ts_l(l)
   
   allocate(b(2))
   allocate(x(2))
   
   b(1)=me2*(dble(1)-ye2)/ye2
   b(2)=(dble(1)-me2)*(dble(1)-ye2)/ye2

   x(1)=e2(l)
   x(2)=dble(1)-e2(l)
   pro = log_DirichletD(x,b,2)
   
   e2(l) = prectdev(xo,d_e2(l),ujsr)
   lln = logl_ts_l(l)
   
   x(1)=e2(l)
   x(2)=dble(1)-e2(l)
   prn = log_DirichletD(x,b,2)
   
   if (dlog(puniformdev(ujsr)).ge.(lln - llo + prn - pro)) then
     e2(l) = xo
     rr_e2(l)=rr_e2(l)+dble(1)
   else
     ar_e2(l)=ar_e2(l)+dble(1)
   endif
   
 end subroutine draw_e2
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine draw_pklm(k,l,ujsr)
   integer, intent(in) :: k,l
   integer, intent(inout) :: ujsr
   integer :: m,ok
   double precision :: sx
   double precision, allocatable :: b(:),x(:)
   
   allocate(b(nall(l)))
   allocate(x(nall(l)))
   
   sx=dble(0)
   if (fk(k)>dble(0)) then
     do m=1,nall(l)
       b(m)=(dble(1)/fk(k)-dble(1))*qlm(l,m)
       x(m)=pgammadev(dble(zklm(k,l,m))+b(m),ujsr)
       sx=sx+x(m)
     enddo
     
     ok=1
     if (sx > dble(1D-100)) then
       do m=1,nall(l)
         x(m)=x(m)/sx
         if (x(m)<dble(1D-100)) ok=0
       enddo

       if (ok==1) then
         do m=1,nall(l)
           pklm(k,l,m)=x(m)
         enddo
       endif
       
     endif
   endif
   
 end subroutine draw_pklm
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine draw_pklmF0(l,ujsr)
   integer, intent(in) :: l
   integer, intent(inout) :: ujsr
   integer :: k,m,ok
   double precision :: sx
   double precision, allocatable :: x(:)
   integer, allocatable :: zlm(:)
   
   allocate(x(nall(l)))
   allocate(zlm(nall(l)))
   
   do m=1,nall(l)
     zlm(m)=0
     do k=1,npop
       zlm(m)=zlm(m)+zklm(k,l,m)
     enddo
   enddo
   
   sx=dble(0)
   do m=1,nall(l)
     x(m)=pgammadev(dble(zlm(m))+dble(1),ujsr)
     sx=sx+x(m)
   enddo
     
   ok=1
   if (sx > dble(1D-100)) then
     do m=1,nall(l)
       x(m)=x(m)/sx
       if (x(m)<dble(1D-100)) ok=0
     enddo
     if (ok==1) then
       do m=1,nall(l)
         do k=1,npop
           pklm(k,l,m)=x(m)
         enddo
       enddo
     endif
   endif
   
 end subroutine draw_pklmF0
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine draw_fk(k,ujsr)
   integer, intent(in) :: k
   integer, intent(inout) :: ujsr
   integer :: l,m
   double precision :: xo,difl,pro,prn,thn,tho
   double precision :: so0,so1,s2,sn0,sn1
   
   pro = log_GammaD(fk(k),mf,kf)
   
   xo=fk(k)
   fk(k)=prectdev(xo,d_fk(k),ujsr)
   
   if (fk(k)<MINFK .or. fk(k)>MAXFK) then
     fk(k)=xo
     return
   endif
   
   thn=dble(1)/fk(k)-dble(1)
   tho=dble(1)/xo-dble(1)
   so0=dble(nl)*DLGAMA(tho)
   so1=dble(0)
   s2=dble(0)
   sn0=dble(nl)*DLGAMA(thn)
   sn1=dble(0)
   do l=1,nl
     do m=1,nall(l)
       so1=so1+DLGAMA(tho*qlm(l,m))
       s2=s2+qlm(l,m)*dlog(pklm(k,l,m))
       sn1=sn1+DLGAMA(thn*qlm(l,m))
     enddo
   enddo
   
   prn = log_GammaD(fk(k),mf,kf)
   
   difl=sn0-sn1+thn*s2-so0+so1-tho*s2
   
   if (dlog(puniformdev(ujsr))>=(difl + prn - pro)) then
     fk(k) = xo
     rr_fk(k)=rr_fk(k)+dble(1)
   else
     ar_fk(k)=ar_fk(k)+dble(1)
   endif
  
 end subroutine draw_fk
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine draw_qlm(j,ujsr)
   integer, intent(in) :: j
   integer, intent(inout) :: ujsr
   integer :: a1,a2,k
   double precision :: xo1,xo2,difl,tk,b1o,b2o,b1n,b2n
   
   if (nall(j)==1) return
   
   a1 = pintuniformdev(nall(j),ujsr) + 1
   do
     a2 = pintuniformdev(nall(j),ujsr) + 1
     if (a1 .ne. a2) exit
   enddo
   
   xo1 = qlm(j,a1)
   xo2 = qlm(j,a2)
   
   qlm(j,a1) = prectdev(xo1,d_qlm(j),ujsr)
   qlm(j,a2) = xo2 - (qlm(j,a1) - xo1)
   
   if (qlm(j,a2) <= MINQLM .or. qlm(j,a2) >= MAXQLM) then
     qlm(j,a1) = xo1
     qlm(j,a2) = xo2
     rr_qlm(j)=rr_qlm(j)+dble(1)
     return
   endif
   
   difl=dble(0)
   do k=1,npop
     tk=(dble(1)/fk(k)-dble(1))
     b1o=tk*xo1
     b2o=tk*xo2
     b1n=tk*qlm(j,a1)
     b2n=tk*qlm(j,a2)
     
     difl=difl+(b1n-b1o)*dlog(pklm(k,j,a1))+(b2n-b2o)*dlog(pklm(k,j,a2)) &
     +DLGAMA(b1o)+DLGAMA(b2o)-DLGAMA(b1n)-DLGAMA(b2n)
   enddo
   
   
   if (dlog(puniformdev(ujsr))>=(difl)) then
     qlm(j,a1) = xo1
     qlm(j,a2) = xo2
     rr_qlm(j)=rr_qlm(j)+dble(1)
   else
     ar_qlm(j)=ar_qlm(j)+dble(1)
   endif 
   
   
 end subroutine draw_qlm
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine draw_y_t
   double precision :: lo,ln,xo,po,pn
   integer :: i
   double precision, allocatable :: bo(:),bn(:),x(:)
   
   allocate(bo(2))
   allocate(bn(2))
   allocate(x(2))
   
   xo=yt
   yt=rectdev(xo,d_yt)
   
   if (yt < MINYT) then
     yt=xo
     rr_yt=rr_yt+dble(1)
     return
   endif
   
   lo=dble(0)
   ln=dble(0)
   po=dble(1)
   pn=dble(1)
   do i=1,nf
     x(1)=ti(i)
     x(2)=dble(1)-ti(i)
     bo(1)=mt(i)*(dble(1)-xo)/xo
     bo(2)=(dble(1)-mt(i))*(dble(1)-xo)/xo
     bn(1)=mt(i)*(dble(1)-yt)/yt
     bn(2)=(dble(1)-mt(i))*(dble(1)-yt)/yt
     lo = lo + log_DirichletD(x,bo,2)
     ln = ln + log_DirichletD(x,bn,2)
   enddo
   
   if (dlog(uniformdev()).ge.(ln - lo + pn - po)) then
     yt = xo
     rr_yt=rr_yt+dble(1)
   else
     ar_yt=ar_yt+dble(1)
   endif
   
 end subroutine draw_y_t
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine draw_my_e1
   double precision :: lo,ln,xo
   integer :: i
   double precision, allocatable :: bo(:),bn(:),x(:)
   
   allocate(bo(2))
   allocate(bn(2))
   allocate(x(2))
   
   xo=me1
   bo(1)=me1*(dble(1)-ye1)/ye1
   bo(2)=(dble(1)-me1)*(dble(1)-ye1)/ye1
   lo=dble(0)
   
   me1=rectdev(xo,d_me1)
   bn(1)=me1*(dble(1)-ye1)/ye1
   bn(2)=(dble(1)-me1)*(dble(1)-ye1)/ye1
   ln=dble(0)
   do i=1,nl
     x(1)=e1(i)
     x(2)=dble(1)-e1(i)
     lo = lo + log_DirichletD(x,bo,2)
     ln = ln + log_DirichletD(x,bn,2)
   enddo
   
   if (dlog(uniformdev()).ge.(ln - lo)) then
     me1 = xo
     rr_me1=rr_me1+dble(1)
   else
     lo=ln
     ar_me1=ar_me1+dble(1)
   endif
   
   xo=ye1
   ye1=rectdev(xo,d_ye1)
   bn(1)=me1*(dble(1)-ye1)/ye1
   bn(2)=(dble(1)-me1)*(dble(1)-ye1)/ye1
   ln=dble(0)
   do i=1,nl
     x(1)=e1(i)
     x(2)=dble(1)-e1(i)
     ln = ln + log_DirichletD(x,bn,2)
   enddo
   
   if (dlog(uniformdev()).ge.(ln - lo)) then
     ye1 = xo
     rr_ye1=rr_ye1+dble(1)
   else
     ar_ye1=ar_ye1+dble(1)
   endif
   
 end subroutine draw_my_e1
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine draw_my_e2
   double precision :: lo,ln,xo
   integer :: i
   double precision, allocatable :: bo(:),bn(:),x(:)
   
   allocate(bo(2))
   allocate(bn(2))
   allocate(x(2))
   
   xo=me2
   bo(1)=me2*(dble(1)-ye2)/ye2
   bo(2)=(dble(1)-me2)*(dble(1)-ye2)/ye2
   lo=dble(0)
   
   me2=rectdev(xo,d_me2)
   bn(1)=me2*(dble(1)-ye2)/ye2
   bn(2)=(dble(1)-me2)*(dble(1)-ye2)/ye2
   ln=dble(0)
   do i=1,nl
     x(1)=e2(i)
     x(2)=dble(1)-e2(i)
     lo = lo + log_DirichletD(x,bo,2)
     ln = ln + log_DirichletD(x,bn,2)
   enddo
   
   if (dlog(uniformdev()).ge.(ln - lo)) then
     me2 = xo
     rr_me2=rr_me2+dble(1)
   else
     lo=ln
     ar_me2=ar_me2+dble(1)
   endif
   
   xo=ye2
   ye2=rectdev(xo,d_ye2)
   bn(1)=me2*(dble(1)-ye2)/ye2
   bn(2)=(dble(1)-me2)*(dble(1)-ye2)/ye2
   ln=dble(0)
   do i=1,nl
     x(1)=e2(i)
     x(2)=dble(1)-e2(i)
     ln = ln + log_DirichletD(x,bn,2)
   enddo
   
   if (dlog(uniformdev()).ge.(ln - lo)) then
     ye2 = xo
     rr_ye2=rr_ye2+dble(1)
   else
     ar_ye2=ar_ye2+dble(1)
   endif
   
 end subroutine draw_my_e2
 !----------------------------------------------------------------------
  
 !----------------------------------------------------------------------
 subroutine draw_mk_f
   double precision :: lo,ln,xo
   double precision :: xn
   integer :: i
   
   xo=mf
   mf=rectdev(xo,d_mf)
   lo=dble(0)
   ln=dble(0)
   do i=1,npop
     lo=lo+log_GammaD(fk(i),xo,kf)
     ln=ln+log_GammaD(fk(i),mf,kf)
   enddo
   
   if (dlog(uniformdev()).ge.(ln-lo)) then
     mf = xo
     rr_mf=rr_mf+dble(1)
   else
     ar_mf=ar_mf+dble(1)
   endif
   
   xo=kf
   xn=rectdev(xo/(dble(1)+xo),d_kf)
   kf=xn/(dble(1)-xn)
   if (kf>dble(9999)) then
     kf=xo
     rr_kf=rr_kf+dble(1)
     return
   endif
   
   lo=dble(0)
   ln=dble(0)
   do i=1,npop
     lo=lo+log_GammaD(fk(i),mf,xo)
     ln=ln+log_GammaD(fk(i),mf,kf)
   enddo
   
   if (dlog(uniformdev()).ge.(ln-lo+dble(2)*(dlog(kf)-dlog(xo)))) then
     kf = xo
     rr_kf=rr_kf+dble(1)
   else
     ar_kf=ar_kf+dble(1)
   endif
   
 end subroutine draw_mk_f
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine draw_beta0
   double precision :: b_o,l_n,l_o,p_o,p_n
   
   l_o = logl_t()
   b_o = beta0;
   p_o = dble(1)
   
   beta0 = b_o + normaldev(dble(0),dsqrt(v_beta0));
   if (beta0<MINB0 .or. beta0>MAXB0) then
     beta0 = b_o
     rr_beta0 = rr_beta0 + dble(1)
     return
   endif
   
   l_n = logl_t()
   p_n = dble(1)

   if (dlog(uniformdev()).ge.(l_n-l_o+p_n-p_o)) then
     beta0 = b_o
     rr_beta0 = rr_beta0 + dble(1)
   else
     ar_beta0 = ar_beta0 + dble(1)
   endif

 end subroutine draw_beta0
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine draw_betas;
   integer :: i 
   double precision :: b_o,l_n,l_o,p_o,p_n
   
   call draw_beta0
   
   if (null_regrmod==0) then
     do i=1,nv
       if (y_beta(i) == 1) then
         l_o = logl_t()
         b_o = beta(i)
         p_o = log_NormD(beta(i),mu_beta,s2_beta)
         
         beta(i) = b_o + normaldev(dble(0),dsqrt(v_beta(i)))
         l_n = logl_t()
         p_n = log_NormD(beta(i),mu_beta,s2_beta)
         if (dlog(uniformdev()).ge.(l_n-l_o+p_n-p_o)) then
           beta(i) = b_o
           rr_beta(i) = rr_beta(i) + dble(1)
         else
           ar_beta(i) = ar_beta(i) + dble(1)
         endif
       endif
     enddo
   endif
 end subroutine draw_betas
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine draw_model_t;
   integer :: m
   double precision :: l_o,l_n,p_o,p_n,b_o
   
   m = intuniformdev(nv)+1

   if (y_beta(m)==1) then
     ! removing a variable
     b_o = beta(m)
     l_o = logl_t()
     p_o = log_NormD(b_o,mu_beta,s2_beta);
     
     y_beta(m) = 0
     l_n = logl_t()
     p_n = log_NormD(dble(0),mq_beta(m),s2q_beta(m))
     
     if (dlog(uniformdev()).ge.(l_n-l_o+p_n-p_o)) then
       y_beta(m) = 1
       l_n = logl_t()
     else 
       beta(m) = dble(0)
     endif
   else
     !adding a variable
     b_o = beta(m)
     l_o = logl_t()
     p_o = log_NormD(b_o,mq_beta(m),s2q_beta(m))
     
     beta(m) = normaldev(mq_beta(m),dsqrt(s2q_beta(m)))
     y_beta(m) = 1
     l_n = logl_t()
     p_n = log_NormD(beta(m),mu_beta,s2_beta)
     
     if (dlog(uniformdev()).ge.(l_n-l_o+p_n-p_o)) then
       y_beta(m) = 0
       beta(m) = 0
       l_n = logl_t()
     endif
   endif
 end subroutine draw_model_t
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine adjust_proposals(fct)
   double precision, intent(in) :: fct
   integer :: i
   double precision :: ar
   
   ar = ar_yt / (ar_yt+rr_yt)
   if (ar < dble(0.3)) then
     d_yt= d_yt/fct
   else
     if (ar > dble(0.5)) then
       d_yt = d_yt*fct
     endif
   endif
   ar_yt = 0
   rr_yt = 0
   
   if (nostruct==0) then
     do i=1,nl
       ar = ar_qlm(i) / (ar_qlm(i)+rr_qlm(i))
       if (ar < dble(0.2)) then
         d_qlm(i) = d_qlm(i)/fct
       else
         if (ar > dble(0.4)) then
           d_qlm(i) = d_qlm(i)*fct
         endif
       endif
       ar_qlm(i) = 0
       rr_qlm(i) = 0
     enddo
   
     do i=1,npop
       ar = ar_fk(i) / (ar_fk(i)+rr_fk(i))
       if (ar < dble(0.2)) then
         d_fk(i) = d_fk(i)/fct
       else
         if (ar > dble(0.4)) then
           d_fk(i) = d_fk(i)*fct
         endif
       endif
       ar_fk(i) = 0
       rr_fk(i) = 0
     enddo
     
     ar = ar_mf / (ar_mf+rr_mf)
     if (ar < dble(0.2)) then
       d_mf = d_mf/fct
     else
       if (ar > dble(0.4)) then
         d_mf = d_mf*fct
       endif
     endif
     ar_mf = 0
     rr_mf = 0
     
     ar = ar_kf / (ar_kf+rr_kf)
     if (ar < dble(0.2)) then
       d_kf = d_kf/fct
     else
       if (ar > dble(0.4)) then
         d_kf = d_kf*fct
       endif
     endif
     ar_kf = 0
     rr_kf = 0
   endif
   
   ar = ar_beta0 / (ar_beta0+rr_beta0)
   if (ar < dble(0.3)) then
     v_beta0 = v_beta0/fct
   else
     if (ar > dble(0.5)) then
       v_beta0 = v_beta0*fct
     endif
   endif
   ar_beta0 = 0
   rr_beta0 = 0
   
   if (null_regrmod==0) then
     do i=1,nv
       ar = ar_beta(i) / (ar_beta(i)+rr_beta(i))
       if (ar < dble(0.3)) then
         v_beta(i) = v_beta(i)/fct
       else
         if (ar > dble(0.5)) then
           v_beta(i) = v_beta(i)*fct
         endif
       endif
       ar_beta(i) = 0
       rr_beta(i) = 0
     enddo
   endif
   
   if (upd_e==1) then
     do i=1,nl
       ar = ar_e1(i) / (ar_e1(i)+rr_e1(i))
       if (ar < dble(0.2)) then
         d_e1(i) = d_e1(i)/fct
       else
         if (ar > dble(0.4)) then
           d_e1(i) = d_e1(i)*fct
         endif
       endif
       ar_e1(i) = 0
       rr_e1(i) = 0
     enddo
     
     ar = ar_me1 / (ar_me1+rr_me1)
     if (ar < dble(0.2)) then
       d_me1 = d_me1/fct
     else
       if (ar > dble(0.4)) then
         d_me1 = d_me1*fct
       endif
     endif
     ar_me1 = 0
     rr_me1 = 0
     
     ar = ar_ye1 / (ar_ye1+rr_ye1)
     if (ar < dble(0.2)) then
       d_ye1 = d_ye1/fct
     else
       if (ar > dble(0.4)) then
         d_ye1 = d_ye1*fct
       endif
     endif
     ar_ye1 = 0
     rr_ye1 = 0
     
     do i=1,nl
       ar = ar_e2(i) / (ar_e2(i)+rr_e2(i))
       if (ar < dble(0.2)) then
         d_e2(i) = d_e2(i)/fct
       else
         if (ar > dble(0.4)) then
           d_e2(i) = d_e2(i)*fct
         endif
       endif
       ar_e2(i) = 0
       rr_e2(i) = 0
     enddo
     
     ar = ar_me2 / (ar_me2+rr_me2)
     if (ar < dble(0.2)) then
       d_me2 = d_me2/fct
     else
       if (ar > dble(0.4)) then
         d_me2 = d_me2*fct
       endif
     endif
     ar_me2 = 0
     rr_me2 = 0
     
     ar = ar_ye2 / (ar_ye2+rr_ye2)
     if (ar < dble(0.2)) then
       d_ye2 = d_ye2/fct
     else
       if (ar > dble(0.4)) then
         d_ye2 = d_ye2*fct
       endif
     endif
     ar_ye2 = 0
     rr_ye2 = 0
   endif
    
 end subroutine adjust_proposals
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine initialize_mcmc
   integer :: i,j,k
   
   !memory allocation follows:
   !-parameters and hyper-parameters
   allocate(qlm(nl,maxl))
   allocate(pklm(npop,nl,maxl))
   allocate(fk(npop))
   allocate(ti(nf))
   allocate(mt(nf))
   allocate(beta(nv))
   allocate(y_beta(nv))
   allocate(e1(nl))
   allocate(e2(nl))
   
   !-parameter means across mcmc
   allocate(mti(nf))
   allocate(mpklm(npop,nl,maxl))
   allocate(mfk(npop))
   allocate(m_e1(nl))
   allocate(m_e2(nl))
   
   !-unobserved data (auxiliary variables)
   allocate(xi(nf))
   allocate(yi(nf))
   allocate(tj(np))
   allocate(zklm(npop,nl,maxl))
   allocate(pgjl(np,nl))
   allocate(xj(np))
   allocate(yj(np))
   
   !proposals and acceptance/rejection rates
   allocate(ar_fk(npop))
   allocate(rr_fk(npop))
   allocate(d_fk(npop))
   allocate(ar_qlm(nl))
   allocate(rr_qlm(nl))
   allocate(d_qlm(nl))
   allocate(v_beta(nv))
   allocate(ar_beta(nv))
   allocate(rr_beta(nv))
   allocate(mq_beta(nv))
   allocate(s2q_beta(nv))
   allocate(d_e1(nl))
   allocate(ar_e1(nl))
   allocate(rr_e1(nl))
   allocate(d_e2(nl))
   allocate(ar_e2(nl))
   allocate(rr_e2(nl))
   
   !initialization follows:
   do j=1,np
     tj(j)=dble(0)
   enddo
   
   do i=1,nl
     do j=1,nall(i)
       qlm(i,j)=dble(1)/dble(nall(i))
     enddo
     e1(i)=ie1
     e2(i)=ie2
     m_e1(i)=  dble(0)
     m_e2(i)=  dble(0)
     d_qlm(i)= dble(0.2)
     ar_qlm(i)=dble(0)
     rr_qlm(i)=dble(0)
     d_e1(i)=  dble(0.2)
     ar_e1(i)= dble(0)
     rr_e1(i)= dble(0)
     d_e2(i)=  dble(0.2)
     ar_e2(i)= dble(0)
     rr_e2(i)= dble(0)
   enddo
   
   do k=1,npop
     do i=1,nl
       do j=1,nall(i)
         pklm(k,i,j)= dble(1)/dble(nall(i))
         mpklm(k,i,j)=dble(0)
       enddo
     enddo
     if (npop>1) then
       fk(k)=ifk*dble(1-nostruct)
     else
       fk(k)= MINFK
     endif
     mfk(k)=  dble(0)
     d_fk(k)= dble(0.2)
     ar_fk(k)=dble(0)
     rr_fk(k)=dble(0)
   enddo
   
   do i=1,nf
     ti(i)=itm
     mti(i)=  dble(0)
   enddo
   
   !beta distributions as priors for outcrossing and error rates
   !parameterized with mean m=a/(a+b) and dispersion y=1/(a+b+1)
   !in order to set uniform priors m should be 0.5 and y should be 1/3
   mt=itm
   yt=   dble(0.3333333)
   d_yt= dble(0.1)
   ar_yt=dble(0)
   rr_yt=dble(0)
   
   me1=ie1
   ye1=   dble(0.3333333)
   d_me1= dble(0.1)
   ar_me1=dble(0)
   rr_me1=dble(0)
   d_ye1= dble(0.1)
   ar_ye1=dble(0)
   rr_ye1=dble(0)
   
   me2=ie2
   ye2=   dble(0.3333333)
   d_me2= dble(0.1)
   ar_me2=dble(0)
   rr_me2=dble(0)
   d_ye2= dble(0.1)
   ar_ye2=dble(0)
   rr_ye2=dble(0)
   
   !gamma distribution for Fst
   !m is the mean value (or about mean, because truncated Gamma distr. is used)
   !k is the shape parameter (k)
   mf=   ifk*dble(1-nostruct)
   kf=   dble(1.00) ! 1.0 degenerates to truncated exponential
   d_mf= dble(0.2)
   ar_mf=dble(0)
   rr_mf=dble(0)
   d_kf= dble(0.5)
   ar_kf=dble(0)
   rr_kf=dble(0)
   
   beta0=  dble(0)
   v_beta0=dble(1)
   do i=1,nv
     beta(i)=    dble(0)
     y_beta(i)=  1-null_regrmod
     v_beta(i)=  dble(0.5)
     mq_beta(i)= dble(0)
     s2q_beta(i)=dble(0)
   enddo
   mu_beta=dble(0)
   s2_beta=dble(10)
   
   if (infer_mgen==1) call init_maternal_genotypes
   logl = full_logl()
   mlogl=dble(0)
   vlogl=dble(0)
   
 end subroutine initialize_mcmc
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine run_mcmc
   !(the main procedure: rj-metropolis-hastings algorithm)
   integer :: u,i,j,k,l,m,nq,ns,s,out_unit1,out_unit2,out_unit3,ut,th,tjsr
   double precision :: mtb0
   character(len=256), allocatable :: pnam(:),fomgi(:)
   integer, allocatable :: out_unit(:)
   integer, allocatable :: akl(:,:)
   character(len=255) :: wdir,mdir
   
   allocate(akl(npop*nl,2))
   i = 0
   do k=1,npop
     do l=1,nl
       i=i+1
       akl(i,1)=k
       akl(i,2)=l
     enddo
   enddo
   
   if (upd_e==1) then
     allocate(pnam(10+nv+nf+npop+2*nl))
   else
     allocate(pnam(6+nv+nf+npop))
   endif
   pnam(1)='LogL'
   pnam(2)='Model'
   pnam(3)='b_0'
   do k=1,nv
     write(pnam(k+3),"(a2,i0)") 'b_',k
   enddo
   pnam(nv+4)='y_t'
   pnam(nv+5)='m_F'
   pnam(nv+6)='k_F'
   do k=1,nf
     write(pnam(nv+6+k),"(a2,i0)") 't_',mfam(k)
   enddo
   do k=1,npop
     write(pnam(nv+nf+6+k),"(a2,i0)") 'F_',popid(k)
   enddo
   
   if (upd_e==1) then
     pnam(nv+nf+npop+7)='m_e1'
     pnam(nv+nf+npop+8)='y_e1'
     do k=1,nl
       write(pnam(nv+nf+npop+8+k),"(a3,i0)") 'e1_',k
     enddo
     
     pnam(nv+nf+npop+nl+9)='m_e2'
     pnam(nv+nf+npop+nl+10)='y_e2'
     do k=1,nl
       write(pnam(nv+nf+npop+10+nl+k),"(a3,i0)") 'e2_',k
     enddo
   endif
   
   u=1
   s=0
   ut=0
   ns=nu/ke
   nq=0
   
   out_unit1 = 20
   out_unit2 = 30
   out_unit3 = 40
   
   open (unit=out_unit1,file=forun,action="write",status="replace")
   if (infer_mgen==1) then
     call getcwd(wdir)
     mdir=domom
     call makedir(wdir,mdir)
     allocate(fomgi(nf))
     allocate(out_unit(nf))
     do i=1,nf
       out_unit(i)=50+i
       write(fomgi(i),"(a,'_',i0,'.mom')") trim(mdir)// dirsep //trim(fomg),mfam(i)
       open (unit=out_unit(i),file=trim(fomgi(i)),action="write",status="replace")
       write (out_unit(i),"(a,*(a,a,i0,a))") 'It',(tab,'L',j,'_a',tab,'L',j,'_b',j=1,nl)
       flush (out_unit(i))
     enddo
   endif
   
   write (out_unit1,"(*(a,a))") 'It',(tab,trim(pnam(k)),k=1,size(pnam))
   
   write(*,"(a)") 'mcmc progress:'
   write(*,"(a)") '              |--------------------+--------------------|'
   write(*,"('pilot tuning:  ',a)", advance="no") char(62)
   
   do i=1,nu+bu+pu
     if (i>bu+pu) ut=1
     
     if (upd_e==1) then
       !$omp parallel num_threads(usernth) default(shared) private(th,tjsr)
       !$omp do
       do j=1,nl
         th=omp_get_thread_num()+1
         tjsr=pjsr(th)
         call draw_e1(j,tjsr)
         pjsr(th)=tjsr
       enddo
       !$omp end do
       !$omp end parallel
       
       !$omp parallel num_threads(usernth) default(shared) private(th,tjsr)
       !$omp do
       do j=1,nl
         th=omp_get_thread_num()+1
         tjsr=pjsr(th)
         call draw_e2(j,tjsr)
         pjsr(th)=tjsr
       enddo
       !$omp end do
       !$omp end parallel
       
       call draw_my_e1
       call draw_my_e2
       
       logl = full_logl()
     endif
     
     if (infer_mgen==1) then
       !$omp parallel num_threads(usernth) default(shared) private(th,tjsr)
       !$omp do
       do j=1,nf
         th=omp_get_thread_num()+1
         tjsr=pjsr(th)
         call draw_Mik(j,tjsr)
         pjsr(th)=tjsr
       enddo
       !$omp end do
       !$omp end parallel
     endif
     
     call infer_aux(ut)
     
     !$omp parallel num_threads(usernth) default(shared) private(th,tjsr)
     !$omp do
     do j=1,nf 
       th=omp_get_thread_num()+1
       tjsr=pjsr(th)
       call draw_ti(j,tjsr)
       pjsr(th)=tjsr
     enddo
     !$omp end do
     !$omp end parallel
     
     call draw_y_t
     
     call draw_betas
     
     if (nostruct==0) then
       !$omp parallel num_threads(usernth) default(shared) private(th,tjsr)
       !$omp do
       do j=1,npop*nl
         th=omp_get_thread_num()+1
         tjsr=pjsr(th)
         call draw_pklm(akl(j,1),akl(j,2),tjsr)
         pjsr(th)=tjsr
       enddo
       !$omp end do
       !$omp end parallel
       
       !$omp parallel num_threads(usernth) default(shared) private(th,tjsr)
       !$omp do
       do j=1,npop
         th=omp_get_thread_num()+1
         tjsr=pjsr(th)
         call draw_fk(j,tjsr)
         pjsr(th)=tjsr
       enddo
       !$omp end do
       !$omp end parallel
       
       call draw_mk_f
       
       !$omp parallel num_threads(usernth) default(shared) private(th,tjsr)
       !$omp do
       do j=1,nl
         th=omp_get_thread_num()+1
         tjsr=pjsr(th)
         call draw_qlm(j,tjsr)
         pjsr(th)=tjsr
       enddo
       !$omp end do
       !$omp end parallel
     else
       !$omp parallel num_threads(usernth) default(shared) private(th,tjsr)
       !$omp do
       do j=1,nl
         th=omp_get_thread_num()+1
         tjsr=pjsr(th)
         call draw_pklmF0(j,tjsr)
         pjsr(th)=tjsr
       enddo
       !$omp end do
       !$omp end parallel
     endif
     
     if (i>bu+pu .and. mod(i-bu-pu,ke)==0) then
       logl = full_logl()
     else
       if (upd_e==1) logl = full_logl()
     endif
     
     if (i<=bu+pu .and. mod(i,1000)==0) call adjust_proposals(dsqrt(dble(2)))
     
     if (i>bu .and. i<=bu+pu .and. mod(i,ke)==0 .and. null_regrmod==0) then
       nq=nq+1
       do j=1,nv
         mq_beta(j)=mq_beta(j)+beta(j)
         s2q_beta(j)=s2q_beta(j)+beta(j)*beta(j)
       enddo
     endif
     
     if (i>bu+pu) then
       if (nv>0 .and. null_regrmod==0 .and. revjump==1) call draw_model_t
       
       if (mod(i-bu-pu,ke)==0) then
         
         mtb0=dble(1)/(dble(1)+dexp(-beta0))
         
         do k=1,npop
           mfk(k)=mfk(k)+fk(k)/dble(ns)
           do l=1,nl
             do m=1,nall(l)
               mpklm(k,l,m)=mpklm(k,l,m)+pklm(k,l,m)/dble(ns)
             enddo
           enddo
         enddo
         if (mod(dble((i-bu-pu)*100)/dble(nu),2.5d0)==0) write(*,"(a)",advance="no") char(62)
         
         mlogl=mlogl+logl/dble(ns)
         vlogl=vlogl+((logl)**2)/dble(ns)
         
         if (upd_e==1) then
         write(out_unit1,"(i0,a,f15.2,a,i0,*(a,f10.5))") i-bu-pu,tab,logl,tab,ytoint(),tab, &
         beta0,(tab,beta(k),k=1,nv),tab,yt,tab,mf,tab,kf,(tab,ti(k),k=1,nf), &
         (tab,fk(k),k=1,npop),tab,me1,tab,ye1,(tab,e1(k),k=1,nl),tab,me2,tab,ye2,(tab,e2(k),k=1,nl)
         else
         write(out_unit1,"(i0,a,f15.2,a,i0,*(a,f10.5))") i-bu-pu,tab,logl,tab,ytoint(),tab, &
         beta0,(tab,beta(k),k=1,nv),tab,yt,tab,mf,tab,kf,(tab,ti(k),k=1,nf),(tab,fk(k),k=1,npop)
         endif
         
         flush(out_unit1)
         
         if (infer_mgen==1) then
           do j=1,nf
             write (out_unit(j),"(i0,*(a,i0,a,i0))") i-bu-pu,(tab,mgen(j,l,1),tab,mgen(j,l,2),l=1,nl)
             flush (out_unit(j))
           enddo
         endif
         
         do j=1,nf
           mti(j)=mti(j)+ti(j)/dble(ns)
         enddo
         
         do l=1,nl
           m_e1(l)=m_e1(l)+e1(l)/dble(ns)
           m_e2(l)=m_e2(l)+e2(l)/dble(ns)
         enddo
         
       endif
     else
       if (mod(i,ke)==0) then
         if (mod(dble(i*100)/dble(bu+pu),2.5d0)==0) write(*,"(a)",advance="no") char(62)
       endif
     endif
     
     if (i==bu+pu) then
       do j=1,nv
       s2q_beta(j)=(s2q_beta(j)-mq_beta(j)*mq_beta(j)/dble(nq))/dble(nq-1)
       mq_beta(j)=mq_beta(j)/dble(nq)
       enddo
       write(*,"(a)",advance="yes") ''
       write(*,"('sampling:      ',a)", advance="no") char(62)
     endif
     
   enddo
   write(*,"(a)",advance="yes") ''
   
   close(out_unit1)
   if (upd_e==1 .and. infer_mgen==1) then
     do j=1,nf
       close(out_unit(j))
     enddo
   endif
   
   !BEGIN: variance of logl
   vlogl=dble(ns)*(vlogl-mlogl**2)/dble(ns-1)
   !END: variance of logl
   
   !BEGIN: summary output file
   open (unit=out_unit3,file=fosum,action="write",status="replace")
   write(out_unit3,"(a)")    '# Data info:'
   write(out_unit3,"(a,a)")  'Data file:             ',trim(fd)
   write(out_unit3,"(a,i0)") 'Number of families:    ',nf
   write(out_unit3,"(a,i0)") 'Number of progeny:     ',np
   write(out_unit3,"(a,i0)") 'Number of loci:        ',nl
   write(out_unit3,"(a,i0)") 'Number of variables:   ',nv
   write(out_unit3,"(a,i0)") 'Number of populations: ',npop
   
   if (nv>0) then
     write(out_unit3,"(a)") ''
     write(out_unit3,"(a)")    '# Mean and Std.Dev. of ecological variables:'
     do j=1,nv
       write(out_unit3,"('Var_',i0,': ',*(f15.4))") j,m_var(j),s_var(j)
     enddo
   endif
   
   write(out_unit3,"(a)") ''
   write(out_unit3,"(a)") '# Parent-offspring incompatibility rates:'
   if (ninc > 0) then
     do l=1,nl
       write(out_unit3,"('Loc_',i0,': ',*(f7.4))") l,pol(l)/ngl(l)
     enddo
   else 
     write(out_unit3,"(a)") 'Zero parent-offspring incompatibilities detected'
   endif
   
   write(out_unit3,"(a)") ''
   write(out_unit3,"(a)")    '# Output info:'
   write(out_unit3,"(a,a)")  'Output file:           ',trim(forun)
   
   write(out_unit3,"(a)") ''
   write(out_unit3,"(a)")    '# MCMC settings:'
   write(out_unit3,"(a,i0)") 'Number of pilot cycles:     ',bu+pu
   write(out_unit3,"(a,i0)") 'Number of samplings cyclse: ',nu
   write(out_unit3,"(a,i0)") 'Number of saved samples:    ',ns
   write(out_unit3,"(a,i0)") 'RNG seed:                   ',seed
   
   if (np>0) then
     write(out_unit3,"(a)") ''
     write(out_unit3,"(a)") '# Mean individual outcrossing rates across MCMC:'
     do i=1,np
       write(out_unit3,"('Ind_',i0,' in Fam_',i0,': ',*(f7.4))") i,mfam(pfam(i)),tj(i)/dble(nu)
     enddo
   endif
   
   if (upd_e==1) then
     write(out_unit3,"(a)") ''
     write(out_unit3,"(a)") '# Mean mistyping error rates across MCMC:'
     do l=1,nl
       write(out_unit3,"(i0,*(f7.4))") l,m_e1(l),e2(l)
     enddo
   endif
   
   write(out_unit3,"(a)") ''
   write(out_unit3,"(a)")      '# Acceptance rates:'
   write(out_unit3,"(a,f5.1,'%')") 'b_0: ',ar_beta0*dble(100)/(ar_beta0+rr_beta0)
   do i=1,nv
     write(out_unit3,"(a,': ',f5.1,'%')") trim(pnam(i+3)),ar_beta(i)*dble(100)/(ar_beta(i)+rr_beta(i))
   enddo
   write(out_unit3,"(a,f5.1,'%')") 'y_t:   ',ar_yt*dble(100) / (ar_yt+rr_yt)
   write(out_unit3,"(a,f5.1,'%')") 'm_F:   ',ar_mf*dble(100) / (ar_mf+rr_mf)
   write(out_unit3,"(a,f5.1,'%')") 'k_F:   ',ar_kf*dble(100) / (ar_kf+rr_kf)
   do i=1,npop
     write(out_unit3,"(a,f5.1,'%')") trim(pnam(nv+nf+6+i)),ar_fk(i)*dble(100)/(ar_fk(i)+rr_fk(i))
   enddo
   do i=1,nl
     write(out_unit3,"('qlm_',i0,':  ',f5.1,'%')") i,ar_qlm(i)*dble(100)/(ar_qlm(i)+rr_qlm(i))
   enddo
   if (upd_e==1) then
     write(out_unit3,"(a,f5.1,'%')") 'm_e1:   ',ar_me1*dble(100) / (ar_me1+rr_me1)
     write(out_unit3,"(a,f5.1,'%')") 'k_e1:   ',ar_ye1*dble(100) / (ar_ye1+rr_ye1)
     do i=1,nl
       write(out_unit3,"('e1_',i0,':  ',f5.1,'%')") i,ar_e1(i)*dble(100)/(ar_e1(i)+rr_e1(i))
     enddo
     write(out_unit3,"(a,f5.1,'%')") 'm_e2:   ',ar_me2*dble(100) / (ar_me2+rr_me2)
     write(out_unit3,"(a,f5.1,'%')") 'k_e2:   ',ar_ye2*dble(100) / (ar_ye2+rr_ye2)
     do i=1,nl
       write(out_unit3,"('e2_',i0,':  ',f5.1,'%')") i,ar_e2(i)*dble(100)/(ar_e2(i)+rr_e2(i))
     enddo
   endif
   
   if (infer_mgen==1) then
     write(out_unit3,"(a)") ''
     write (out_unit3,"(a)") '# Allele coding for maternal genotypes:'
     write (out_unit3,"(a,*(a,a,i0))") 'Code',(tab,'Loc_',i,i=1,nl)
     do i=1,ubound(allele,2)
       write(out_unit3,"(i0,*(a,i0))") i,(tab,allele(j,i),j=1,nl)
     enddo
   endif
   
   write(out_unit3,"(a)") ''
   write(out_unit3,"(a)")    '# Program info:'
   write(out_unit3,"(a,a)")  'Version number:  ',trim(VERSION)
   write(out_unit3,"(a,a)")  'Date of realese: ',trim(LASTUPDATE)
   write(out_unit3,"(a,a)")  'Compiler version: ',compiler_version()
   
   close(out_unit3)
   !END: summary output file
   
 end subroutine run_mcmc
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function rgamma(a)
   double precision, intent(in) :: a
   double precision :: x,v,u,c,d
   
   d=a-dble(1)/dble(3)
   c=dble(1)/dsqrt(dble(9)*d)
   
   do
     do
       x=dsqrt(-dble(2)*dlog(uniformdev()))*dcos(my2pi*uniformdev())
       v=(dble(1)+c*x)
       if (v > dble(0)) exit
     enddo
     
     v=v**3
     u=uniformdev()
     
     if (u<(dble(1)-dble(0.0331)*x**4)) exit
     if (dlog(u)<(dble(0.5)*x**2+d*(dble(1)-v+dlog(v)))) exit
   enddo
   
   rgamma=d*v
 end function rgamma
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function prgamma(a,ujsr)
   double precision, intent(in) :: a
   integer, intent(inout) :: ujsr
   double precision :: x,v,u,c,d
   
   d=a-dble(1)/dble(3)
   c=dble(1)/dsqrt(dble(9)*d)
   
   do
     do
       x=dsqrt(-dble(2)*dlog(puniformdev(ujsr)))*dcos(my2pi*puniformdev(ujsr))
       v=(dble(1)+c*x)
       if (v > dble(0)) exit
     enddo
     
     v=v**3
     u=puniformdev(ujsr)
     
     if (u<(dble(1)-dble(0.0331)*x**4)) exit
     if (dlog(u)<(dble(0.5)*x**2+d*(dble(1)-v+dlog(v)))) exit
   enddo
   
   prgamma=d*v
 end function prgamma
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function uniziggurat()
   integer :: jz,shr3
   
   jz = jsr
   jsr = ieor( jsr, ishft( jsr,  13 ) )
   jsr = ieor( jsr, ishft( jsr, -17 ) )
   jsr = ieor( jsr, ishft( jsr,   5 ) )
   shr3 = jz + jsr
   
   uniziggurat = dble(0.5) + 0.2328306D-9 * dble(shr3)
 end function uniziggurat
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function puniziggurat(ujsr)
   integer, intent(inout) :: ujsr
   integer :: jz,shr3
   
   jz = ujsr
   ujsr = ieor( ujsr, ishft( ujsr,  13 ) )
   ujsr = ieor( ujsr, ishft( ujsr, -17 ) )
   ujsr = ieor( ujsr, ishft( ujsr,   5 ) )
   shr3 = jz + ujsr
   
   puniziggurat = dble(0.5) + 0.2328306D-9 * dble(shr3)
 end function puniziggurat
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 integer function ytoint()
   integer :: m,r
   
   r=0
   do m=1,nv
     r=r+y_beta(m)*(2**(m-1))
   enddo
   ytoint = r
   
 end function ytoint
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine makedir(path,rdir)
   character(len=255), intent(in) :: path
   character(len=255), intent(inout) :: rdir
   character(len=255) :: makedirectory,checkdir
   logical :: exists
   
   makedirectory = 'mkdir "' // trim(path) // dirsep // trim(rdir) // '"'
   checkdir = trim(path) // dirsep // trim(rdir) // dirsep // '.'
   inquire( file = checkdir, exist = exists )
   if (.not. exists) call system(makedirectory)
   rdir = trim(path) // dirsep // trim(rdir)
   
 end subroutine makedir
 !----------------------------------------------------------------------
 
end program quedo
